from lxml.etree import QName
import re
import keyword


def avoid_keyword(name: str):
    if keyword.iskeyword(name) or name == "print":
        return "_" + name
    return name


def get_doc(callable_tag):
    """Return docstring text for a callable"""
    for element in callable_tag:
        tag = QName(element)
        if tag.localname == "doc":
            return str(element.text).replace("\\x", "x")
    return ""


def make_safe(string):
    """Avoid having unicode characters in docstrings (such as uXXXX)"""
    return string.replace("\\u", "u").replace("\\U", "U")


def prettify(string):
    return re.sub(r"([\s]{3,80})", r"\n\1", string)


GIR_TO_NATIVE_TYPEMAP = {
    "gboolean": "bool",
    "gint": "int",
    "guint": "int",
    "gint8": "int",
    "guint8": "int",
    "gint16": "int",
    "guint16": "int",
    "gint32": "int",
    "guint32": "int",
    "gint64": "int",
    "guint64": "int",
    "gsize": "int",
    "gpointer": "typing.Any",
    "none": "None",
    "gchar": "str",
    "guchar": "str",
    "gchar*": "str",
    "guchar*": "str",
    "glong": "long",
    "gulong": "long",
    "glong64": "long",
    "gulong64": "long",
    "gshort": "int",
    "gushort": "int",
    "gshort64": "int",
    "gushort64": "int",
    "gfloat": "float",
    "gdouble": "float",
    "string": "str",
    "GString": "str",
    "utf8": "str",
}


def get_native_type(typename: str):
    """Convert a C type to a Python type"""
    typename = typename.replace("const ", "")
    return GIR_TO_NATIVE_TYPEMAP.get(typename, typename)


def get_type_from_node(element):
    for child in element:
        subtag = QName(child)
        if subtag.localname == "type":
            try:
                return get_native_type(child.attrib["name"]) or "typing.Any"
            except KeyError:
                continue

    return "typing.Any"


def get_imports_from_type(typ: str):
    return typ[: typ.index(".")] if "." in typ else None


def get_returntype(element):
    """Return the return-type of a callable"""
    for elem_property in element:
        tag = QName(elem_property)
        if tag.localname == "return-value":
            return_doc = get_doc(elem_property).replace("\n", " ").strip()
            return_type = get_type_from_node(elem_property)
            return (return_doc, return_type)
    return ("", "None")



from pathlib import Path
from typing import Optional
import typer
import os
import sys
import site

from gengir.main import generate_fakegir


def cli(
    types: Optional[list[Path]] = typer.Argument(
        None,
        help="Files to use as input for the generator. If not provided it uses all files in /usr/share/gir-1.0/",
    ),
    outdir: Optional[Path] = typer.Option(
        None,
        "--outdir",
        "-o",
        help="Directory to store the package typings. $site-packages/gi-stubs by default",
    ),
    docs: bool = typer.Option(True, help="Include docstrings in the typings"),
    gtk: int = typer.Option(3, help="GTK version to generate typings for"),
):
    """
    Generate PEP 561 stubs for the GObject introspection library.
    """

    sp = Path(site.getsitepackages()[0])
    if not os.access(sp, os.W_OK | os.X_OK):
        user_sp = Path(site.getusersitepackages())
        print(f"Can't write to {sp}, using {user_sp} instead", file=sys.stderr)
        sp = user_sp

    outdir = outdir or sp / "gi-stubs"

    generate_fakegir(
        {
            "files": types or [],
            "gtk_version": gtk,
            "include_docs": docs,
            "outdir": outdir,
        }
    )


def run_cli():
    typer.run(cli)



#!/usr/bin/env python3
"""Build a fake python package from the information found in gir files"""
import glob
import os
from pathlib import Path
import multiprocessing
import logging
from typing import Optional, TypedDict
import astor

from lxml.etree import XML, XMLParser

from gengir.generate import StubGenerator

logger = logging.getLogger(__name__)

XMLNS = "http://www.gtk.org/introspection/core/1.0"


class Options(TypedDict):
    files: list[Path]
    outdir: Path
    include_docs: bool
    gtk_version: int


GIR_PATH = Path("/usr/share/gir-1.0")


def iter_girs(opt: Options):
    files = opt["files"]
    gtk_version = opt["gtk_version"]

    """Return a generator of all available gir files"""
    gir_files = files if len(files) > 0 else GIR_PATH.glob("*.gir")

    for gir_file in gir_files:
        basename = gir_file.name

        if len(files) == 0:
            # Check which GTK Version the user wants to use
            if (
                (
                    basename in ("Gtk-2.0.gir", "Gdk-2.0.gir", "GdkX11-2.0.gir")
                    and gtk_version != 2
                )
                or (
                    basename in ("Gtk-3.0.gir", "Gdk-3.0.gir", "GdkX11-3.0.gir")
                    and gtk_version != 3
                )
                or (
                    basename in ("Gtk-4.0.gir", "Gdk-4.0.gir", "GdkX11-4.0.gir")
                    and gtk_version != 4
                )
            ):
                continue

        try:
            module_name = basename[: basename.index("-")]
        except ValueError:
            # file name contains no dashes
            logger.warning(f"Warning: unrecognized file in gir directory: {gir_file}")
            continue

        yield (opt, module_name, gir_file)


def generate_module(gir_info: tuple[Options, str, Path]):
    opt, module_name, gir_path = gir_info

    print("Parsing {}".format(gir_path))
    parser = XMLParser(encoding="utf-8", recover=True)
    with open(gir_path, "rt", encoding="utf-8") as gir_file:
        content = gir_file.read()
    root = XML(content, parser)
    namespace = root.findall("{%s}namespace" % XMLNS)[0]

    gen = StubGenerator(include_docs=opt["include_docs"])

    namespace_content = astor.to_source(gen.extract_namespace(namespace))

    fakegir_path = opt["outdir"] / "repository" / (module_name + ".pyi")

    with open(fakegir_path, "w") as fakegir_file:
        fakegir_file.write("# -*- coding: utf-8 -*-\n")
        fakegir_file.write(namespace_content)


def generate_fakegir(opt: Options):
    """Main function"""

    os.makedirs(opt["outdir"] / "repository", exist_ok=True)

    with open(opt["outdir"] / "__init__.pyi", "w") as init:
        init.write("def require_version(namespace, version):\n")
        init.write("    pass\n")

    with open(opt["outdir"] / "repository/__init__.py", "w") as repo_init:
        repo_init.write("")

    pool = multiprocessing.Pool()

    pool.map(generate_module, iter_girs(opt))



import ast
from typing import TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from lxml.etree import _Element

from gengir.util import (
    avoid_keyword,
    get_doc,
    get_returntype,
    make_safe,
    prettify,
    get_imports_from_type,
    get_type_from_node,
)

from lxml.etree import QName

XMLNS = "http://www.gtk.org/introspection/core/1.0"


class Param(TypedDict):
    name: str
    typ: str
    doc: str
    variadic: bool


class StubGenerator:
    imports = set[str]()

    def __init__(self, *, include_docs: bool = True) -> None:
        self.include_docs = include_docs

    def imports_from_type(self, typ: str):
        if typ := get_imports_from_type(typ):
            self.imports.add(typ)

    def get_parameter_doc(self, element):
        """Returns the doc of a parameter"""
        param_doc = ""
        if self.include_docs:
            for elem_property in element:
                tag = QName(elem_property)
                if tag.localname == "doc":
                    param_doc = (
                        element.text.replace("\\x", "x")
                        .encode("utf-8")
                        .replace("\n", " ")
                        .strip()
                    )
                    break

        return param_doc

    def get_parameters(self, element):
        """Return the parameters of a callable"""
        args = list[Param]()

        for elem_property in element:
            tag = QName(elem_property)
            if tag.localname == "parameters":
                for param in elem_property:
                    subtag = QName(param)
                    param_name: str

                    if subtag.localname == "instance-parameter":
                        param_name = "self"
                    else:
                        try:
                            param_name = param.attrib["name"]
                        except KeyError:
                            continue

                    param_name = avoid_keyword(param_name)
                    param_type = get_type_from_node(param)
                    param_doc = make_safe(get_doc(param).replace("\n", " ").strip())

                    is_variadic = param_name == "..."
                    if is_variadic:
                        param_name = "args"

                    if param_name not in [p["name"] for p in args]:
                        self.imports_from_type(param_type)

                        args.append(
                            Param(
                                name=param_name,
                                doc=param_doc,
                                typ=param_type,
                                variadic=is_variadic,
                            )
                        )

        return args

    def make_function_ast(
        self,
        name: str,
        parameters: list[Param],
        returntype: tuple[str, str],
        docstring: str,
        static: bool = False,
    ):
        """Returns a function as a AST node"""
        fun = ast.FunctionDef()
        fun.name = avoid_keyword(name)

        fun.decorator_list = []
        if static:
            fun.decorator_list.append(ast.Name("staticmethod"))

        fun.args = ast.arguments(
            posonlyargs=[],
            args=[],
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        )
        fun.returns = ast.alias(returntype[1])

        docstring += "\n\n"

        for param in parameters:
            if param["variadic"]:
                fun.args.vararg = ast.arg("args", ast.alias(param["typ"]))
            elif param["name"] == "self":
                fun.args.args.append(ast.arg("self"))
            else:
                fun.args.args.append(ast.arg(param["name"], param["typ"]))
                fun.args.defaults.append(ast.Constant(None))

            if self.include_docs:
                docstring += ":param {name}: {doc}\n".format(**param)

        if self.include_docs:
            docstring += ":return: {}".format(prettify(returntype[0]))

        self.imports_from_type(returntype[1])

        fun.body = [ast.Expr(ast.Constant(docstring)), ast.Pass()]

        return fun

    def extract_methods(self, class_tag):
        """Return methods from a class element"""

        methods = list[ast.FunctionDef]()

        for element in class_tag:
            tag = QName(element)
            if tag.localname in ("function", "method", "virtual-method", "constructor"):
                method_name: str = element.attrib["name"]
                if method_name == "print":
                    method_name += "_"

                docstring = get_doc(element)
                params = self.get_parameters(element)

                decoration = (
                    "" if any(p["name"] == "self" for p in params) else "staticmethod"
                )
                returntype = get_returntype(element)

                if tag.localname == "constructor" and method_name == "new":
                    params_init = list(params)
                    params_init.insert(0, Param(name="self", doc="", variadic=False))
                    methods.append(
                        self.make_function_ast(
                            "__init__",
                            [Param(name="self", doc="", variadic=False), *params],
                            ("", "None"),
                            docstring,
                        )
                    )

                methods.append(
                    self.make_function_ast(
                        method_name, params, returntype, docstring, decoration
                    )
                )
        return methods

    def extract_fields(self, record_tag):
        """Return fields from a record element"""
        fields_content = list[ast.AnnAssign]()

        for element in record_tag:
            tag = QName(element)
            if tag.localname == "field":
                field_name = element.attrib["name"]
                field_type = get_type_from_node(element)
                self.imports_from_type(field_type)

                fields_content.append(
                    ast.AnnAssign(
                        target=ast.Name(field_name),
                        annotation=ast.alias(field_type),
                        simple=1,
                    )
                )
        return fields_content

    def get_bases(self, klass: ast.ClassDef):
        parents = list[str]()
        for base in klass.bases:
            parents.append(ast.unparse(base))
        return parents

    def build_classes(self, classes: list[ast.ClassDef]):
        """Order classes with correct dependency order
        also return external imports
        """
        ordered_classes = list[ast.ClassDef]()

        local_parents = set[str]()
        written_classes = set[str]()
        all_classes = set[str]([klass.name for klass in classes])

        for klass in classes:
            parents = self.get_bases(klass)
            local_parents = local_parents.union(
                set(
                    [
                        class_parent
                        for class_parent in parents
                        if "." not in class_parent
                    ]
                )
            )

        while written_classes != all_classes:
            for klass in classes:
                class_name, parents = (
                    klass.name,
                    self.get_bases(klass),
                )
                skip = False
                for parent in parents:
                    self.imports_from_type(parent)
                    if "." not in parent and parent not in written_classes:
                        skip = True
                if class_name in written_classes:
                    skip = True

                if skip:
                    continue

                ordered_classes.append(klass)

                written_classes.add(class_name)

        return ordered_classes

    def extract_class(self, element):
        """Extract information from a class"""

        klass = ast.ClassDef()
        klass.name = element.attrib["name"]
        klass.bases = []
        klass.body = []
        klass.decorator_list = []

        docstring = get_doc(element)
        klass.body.append(ast.Expr(ast.Constant(docstring)))

        if parent := element.attrib.get("parent"):
            self.imports_from_type(parent)
            klass.bases.append(ast.alias(parent))

        implements = element.findall("{%s}implements" % XMLNS)
        for implement in implements:
            name = implement.attrib["name"]
            self.imports_from_type(name)
            klass.bases.append(ast.alias(name))

        methods = self.extract_methods(element)
        fields = self.extract_fields(element)

        klass.body.extend([*methods, *fields])

        return klass

    def insert_enum(self, element: "_Element"):
        """Returns an enum (class with attributes only) as text"""
        klass = ast.ClassDef()
        klass.name = element.attrib["name"]
        klass.bases = [ast.alias("enum.Enum")]
        klass.body = []
        klass.decorator_list = []

        docstring = get_doc(element)
        klass.body.append(ast.Expr(ast.Constant(docstring)))

        members = element.findall("{%s}member" % XMLNS)
        for member in members:
            const_name: str = member.attrib["name"]
            if not const_name:
                const_name = "_"
            if const_name and const_name[0].isdigit():
                const_name = "_" + const_name
            const_value: str = member.attrib["value"]
            const_value = const_value.replace("\\", "\\\\")
            if const_value == "(null)":
                const_value = 0
            klass.body.append(
                ast.Assign(
                    targets=[ast.Name(const_name.upper())],
                    value=ast.Constant(int(const_value)),
                )
            )

        return klass

    def extract_namespace(self, namespace: "_Element"):
        """Extract all information from a gir namespace"""

        namespace_content = list[ast.stmt]()
        classes = list[ast.ClassDef]()
        for element in namespace:
            tag = QName(element)
            tag_name = tag.localname

            if tag_name in ("class", "interface", "record"):
                klass = self.extract_class(element)
                classes.append(klass)

            elif tag_name in ("enumeration", "bitfield"):
                self.imports.add("enum")
                namespace_content.append(self.insert_enum(element))

            elif tag_name == "function":
                params = self.get_parameters(element)
                returns = get_returntype(element)
                self.imports_from_type(returns)

                namespace_content.append(
                    self.make_function_ast(
                        element.attrib["name"],
                        params,
                        returns,
                        get_doc(element),
                    )
                )
            elif tag_name == "constant":
                constant_name = element.attrib["name"]
                if constant_name[0].isdigit():
                    constant_name = "_" + constant_name
                constant_value = element.attrib["value"] or ""
                constant_value = constant_value.replace("\\", "\\\\")
                namespace_content.append(
                    ast.Assign(
                        targets=[ast.Name(constant_name)],
                        value=ast.Constant(constant_value),
                    )
                )

        classes_content = self.build_classes(classes)
        namespace_content.extend(classes_content)

        for _import in self.imports:
            imp: ast.stmt

            if not _import:
                continue
            elif _import in ("typing", "enum"):
                imp = ast.Import([ast.alias(_import)])
            else:
                imp = ast.ImportFrom("gi.repository", [ast.alias(_import)], 0)
            namespace_content.insert(0, imp)

        return ast.Module(body=namespace_content)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gengir']

package_data = \
{'': ['*']}

install_requires = \
['astor>=0.8.1,<0.9.0', 'lxml>=4.6.4,<5.0.0', 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['gengir = gengir.cli:run_cli']}

setup_kwargs = {
    'name': 'gengir',
    'version': '1.0.1',
    'description': 'Generate PEP 561 stubs for the GObject introspection library. Forked from fakegir',
    'long_description': "# GenGIR: Genuine* autocompletion for your PyGObject code!\n\n\n[fakegir](https://github.com/strycore/fakegir) is a tool to build a fake python package of PyGObject modules.\n\nGenGIR is a fork of fakegir that uses modern standards and improves on usability \n\nThe main changes are:\n\n-   Use [PEP 484](https://www.python.org/dev/peps/pep-0484/) type annotations instead of docstrings\n-   Install typings as a [PEP 561](https://www.python.org/dev/peps/pep-0561/) stub\n-   Use Sphinx markup on docstrings\n-   A cli\n\nGenGIR stores the type info inside your user or venv site-packages as a package named `gi-stubs`.  \nOnce it's installed, it should be recognized by your IDE and it should provide autocompletion and typing errors.\n\n\n## Building\n\nThis project uses [poetry](https://python-poetry.org/), so make sure to have that installed\n\nThen run `poetry install` and `poetry build`. A wheel file should be created in the dist directory that can be installed using `pip install ./dist/gengir*.whl`\n\n\n## Usage\n\nThe `*.gir` with the type info files should be included with each GNOME library development package in `/usr/share/gir-1.0/`.\n\n```\ngengir [OPTIONS] [TYPES]...\n\n  Generate PEP 561 stubs for the GObject introspection library.\n\nArguments:\n  [TYPES]...  Files to use as input for the generator. If not provided it uses\n              all files in /usr/share/gir-1.0/\n\nOptions:\n  -o, --outdir PATH               Directory to store the package typings.\n                                  $site-packages/gi-stubs by default\n  --docs / --no-docs              Include docstrings in the typings  [default:\n                                  docs]\n  --gtk INTEGER                   GTK version to generate typings for\n                                  [default: 3]\n```\n\n## Editor support\n\n-   VSCode has support for stub packages out of the box.\n-   [Jedi](https://github.com/davidhalter/jedi) supports it too, so any editor using it should work.\n",
    'author': 'Santiago Cézar',
    'author_email': 'santiagocezar2013@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/santiagocezar/gengir',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
